from torch.utils.tensorboard import SummaryWriter

from yangdl.core.env import env
from yangdl.metric import Metric


__all__ = [
    'TensorBoardLogger',
]


class TensorBoardLogger():
    """The logger of tensorboard."""
    def __init__(self, fold: int):
        if env.exp_path is not None:
            self.writer = SummaryWriter(f'{env.exp_path}/log/{fold}')
        else:
            self.writer = None

    def log(
        self, 
        metric: Metric,
        stage: str,
        step: int,
    ):
        """Log `Metric` to tensorboard."""
        if self.writer is None:
            return

        for prop_name, prop in metric.data.items():
            if isinstance(prop, list): # lists are not currently logged
                continue

            if metric.prefix is not None:
                prop_name = f'{metric.prefix}_{prop_name}'
            self.writer.add_scalar(f'{prop_name}/{stage}', prop, step)

    def __del__(self):
        if self.writer is not None:
            self.writer.close()




from yangdl.core.model_module import ModelModule
from yangdl.core.data_module import DataModule
from yangdl.core.task_module import TaskModule
from yangdl.core.logger import logger
from yangdl.core.env import env


__all__ = [
    'ModelModule',
    'DataModule',
    'TaskModule',
    'logger',
    'env',
]



from rich.progress import (
    Progress,
    BarColumn,
    MofNCompleteColumn,
)


__all__ = [
    'EpochProgress',
]


class EpochProgress(Progress):
    """Display the experimental process in the form of a progress bar."""
    def __init__(self, stage: str, total: int, fold: int, epoch: int):
        super().__init__(transient=True)

        self.columns = (
            '{task.description}: fold: {task.fields[fold]}, epoch: {task.fields[epoch]}',
            BarColumn(),
            MofNCompleteColumn(),
        )
        self.keys = set()
        self.color = {'train': 'red', 'val': 'blue', 'test': 'green', 'predict': 'pink'}[stage]
        self.task_id = self.add_task(description=f'[{self.color}]{stage}[/{self.color}]', total=total, fold=fold, epoch=epoch)

    def update(self, **kwargs):
        """Display the `**kwargs` behind the progress bar."""
        for key in kwargs:
            if key not in self.keys:
                self.keys.add(key)
                self.columns = (*self.columns, f'{key}: [{self.color}]{{task.fields[{key}]:.3f}}[/{self.color}]')

        super().update(self.task_id, **kwargs)
        super().advance(self.task_id)




import torch
from torch import Tensor

from yangdl.metric import Metric


class ValueMetric(Metric):
    """Accumulate a single value."""
    def __init__(
        self,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.add_tensor('_val', torch.tensor(0, dtype=torch.float))
        self.add_tensor('n', torch.tensor(0, dtype=torch.long))

        self.properties = ['val']

    @torch.no_grad()
    def update(self, val: Tensor, n: int = 1):
        """Update metric with `n` numbers whose average is `val`.

        A common use case is to accumulate loss.
        In this case, `val` is average loss of the entire batch,
        and n is the number of samples in the batch.

        Args:
            val: Average of data.
            n: Length of data.
        """
        val = val.detach()
        if isinstance(val, Tensor):
            val = val.item()

        self._val += val * n
        self.n += n

    @property
    def val(self) -> Tensor:
        if self.n == 0:
            return torch.tensor(-1.)
        return self._val / self.n



import json
import os


__all__ = [
    'dict2json',
]


def dict2json(file_name: str, d: dict) -> None:
    """Write dict to json.

    Args:
        file_name: The file name to write dict.
        d: Data dict.
    """
    assert file_name[-4:] == 'json'

    class MyEncoder(json.JSONEncoder):
        def default(self, obj):
            return str(obj)

    os.makedirs(os.path.dirname(file_name), exist_ok=True)
    with open(file_name, 'w') as f:
        f.write(json.dumps(d, indent=2, separators=(', ', ': '), cls=MyEncoder))




import logging

from rich.logging import RichHandler
import torch

from yangdl.core.env import env
from yangdl.utils.helper import clear_markup


__all__ = [
    'logger',
]


class CombinationHandler(logging.Handler):
    """`RichHandler` for logging to the terminal and `FileHandler` for outputting to the file."""
    def __init__(self, file_name: str = '/dev/null'):
        super().__init__()

        self.rich_handler = RichHandler(
            level=logging.NOTSET,
            show_path=False,
            markup=True,
            rich_tracebacks=True,
            log_time_format='%H:%M:%S'
        )

        self.file_handler = logging.FileHandler(filename=file_name, mode='w', delay=True)
        self.file_handler.setFormatter(logging.Formatter(
            fmt='%(asctime)s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        self.file_handler.setLevel(logging.NOTSET)

    def emit(self, record):
        self.rich_handler.emit(record)
        if self.file_handler.baseFilename != '/dev/null':
            record.msg = clear_markup(record.msg)
            self.file_handler.emit(record)

    def set_file_name(self, file_name: str) -> None:
        self.file_handler.baseFilename = file_name


class Logger:
    """Encapsulates logging module (Use singleton `logger` instead of instantiating `Logger`).

    Mainly used to print and record logs.
    """
    def __init__(self, file_name: str = '/dev/null'):
        """
        Args:
            file_name: File to save logs.
        """
        self.logger = logging.getLogger('root')
        self.logger.setLevel(logging.NOTSET)

        self.handler = CombinationHandler(file_name)
        self.logger.addHandler(self.handler)

    def info(self, msg: str, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs, extra={'markup': True})

    def log_props(self, fmt='{:.3f}', **kwargs):
        """Log properties.

        Examples:
            >>> logger.log_props(loss=self.loss.val, auc=self.metric.auc)
            {stage}: fold: {fold}, epoch: {epoch}: loss: 0.3, auc: 0.7
        """
        stage, fold, epoch = env.stage, env.fold, env.epoch
        color = {'train': 'red', 'val': 'blue', 'test': 'green', 'predict': 'pink'}[stage]

        s = f'[{color}]{stage:>5}[/{color}]: fold:{fold:>2}, epoch: {epoch:>2}: '
        for key, val in kwargs.items():
            if isinstance(val, torch.Tensor):
                val = val.cpu().item()
            s += f'{key}: [{color}]{fmt.format(val)}[/{color}], '
        s = s[:-2]

        self.logger.info(s, extra={'markup': True, 'highlighter': None})


logger = Logger()




import re


class WithNone():
    def __enter__(self):
        pass

    def __exit__(self, err_type, err_val, err_pos):
        pass


def clear_markup(s: str):
    """
    Remove square brackets and their contents.
    """
    return re.sub(r'(\[.*?\])', '', s)


def apply_format_to_float(val, fmt: str):
    if isinstance(val, float):
        return fmt.format(val)
    elif isinstance(val, dict):
        return {k: apply_format_to_float(v, fmt) for k, v in val.items()}
    elif isinstance(val, list):
        return [apply_format_to_float(v, fmt) for v in val]
    else:
        return val



from collections import defaultdict
import copy
import datetime
import json
import os
from typing import Optional

import numpy as np
import torch
from torch import Tensor
from torch.cuda.amp import autocast
from torch.utils.data import DataLoader

from yangdl.core.model_module import ModelModule
from yangdl.core.data_module import DataModule
from yangdl.core.logger import logger
from yangdl.core.env import env
from yangdl.core.early_stop import EarlyStop
from yangdl.core.tensorboard_logger import TensorBoardLogger
from yangdl.core.ui import EpochProgress
from yangdl.metric import Metric
from yangdl.utils.helper import (
    WithNone,
    apply_format_to_float,
)
from yangdl.utils.io import dict2json
from yangdl.utils.python import method_is_overrided_in_subclass


__all__ = [
    'TaskModule',
]


class TaskModule():
    """Encapsulate the basic process of deep learning.

    You need to pass in parameters to instantiate `TaskModule`,
    and then just run the `do` function.

    Examples:
        >>> task_module = TaskModule(model_module=MyModelModule(), data_module=MyDataModule())
        >>> res = task_module.do()
        {'train': {'metric': {'acc': 0.7, 'auc': 0.8}}}
    """
    def __init__(
        self,
        model_module: ModelModule,
        data_module: DataModule,
        early_stop_params: Optional[dict] = None,
        save_ckpt_period: int = 0,
        val_first: bool = False,
        fmt: str = '{:.4f}',
        benchmark: bool = False,
        deterministic: bool = True,
    ):
        """
        Args:
            model_module: Inherited from `yangdl.ModelModule`.
            data_module: Inherited from `yangdl.DataModule`.
            early_stop_params: Is a params dict to initialize `EarlyStop`. Refer to `EarlyStop` for more details.
            save_ckpt_period: Checkpoints will be saved every `save_ckpt_period` epochs.
                If equal to 0, only `best.pt` will be saved.
            val_first: If set to True, before train at 1 epoch, val will be at 0 epoch.
            fmt: Control the format of output metric values.
            benchmark: Refer to `torch.backends.cudnn.benchmark` for more information.
            deterministic: Refer to `torch.backends.cudnn.deterministic` for more information.
        """

        self.model_module = model_module
        self.data_module = data_module
        self.early_stop = EarlyStop(**early_stop_params) if early_stop_params is not None else None
        self.tensorboard_logger = None
        self.save_ckpt_period = save_ckpt_period
        self.val_first = val_first
        self.fmt = fmt
        torch.backends.cudnn.benchmark = benchmark
        torch.backends.cudnn.deterministic = deterministic

        self.train: bool = method_is_overrided_in_subclass(data_module.train_loader)
        self.val: bool = method_is_overrided_in_subclass(data_module.val_loader)
        self.test: bool = method_is_overrided_in_subclass(data_module.test_loader)
        self.predict: bool = method_is_overrided_in_subclass(data_module.predict_loader)

        self.train_named_metrics: list[dict[str, Metric]] = []
        self.val_named_metrics: list[dict[str, Metric]] = []
        self.test_named_metrics: list[dict[str, Metric]] = []

    @staticmethod
    def _log_wrapper(func):
        """Log some information for `do` method."""
        def wrapper(self, *args, **kwargs):
            start = datetime.datetime.now()
            res = func(self, *args, **kwargs)
            logger.info(json.dumps(apply_format_to_float(res, self.fmt), indent=2))
            s = (datetime.datetime.now() - start).total_seconds()
            h = s // 3600
            s = s % 3600
            m = s // 60
            s = s % 60
            logger.info('Total time: %d:%02d:%02d', h, m, s)
            logger.info('Experiment path: %s', env.exp_path)
            return res

        return wrapper

    @_log_wrapper
    def do(self) -> None:
        """Begin your multi-fold train, val, test, predict!"""
        for fold, (_, train_loader, val_loader, test_loader, predict_loader) in enumerate(zip(self.model_module, self.data_module.train_loader(), self.data_module.val_loader(), self.data_module.test_loader(), self.data_module.predict_loader()), 1):
            self.model_module.cuda()
            self.tensorboard_logger = TensorBoardLogger(fold)
            if self.early_stop is not None:
                self.early_stop.init()

            self._do_fold(fold, train_loader, val_loader, test_loader, predict_loader)

        metric_path = f'{env.exp_path}/metric'
        res = {}
        if self.train:
            train_res = self._named_metrics2res(self.train_named_metrics)
            dict2json(apply_format_to_float(f'{metric_path}/train.json', self.fmt), train_res)
            res['train'] = train_res
        if self.val:
            val_res = self._named_metrics2res(self.val_named_metrics)
            dict2json(apply_format_to_float(f'{metric_path}/val.json', self.fmt), val_res)
            res['val'] = val_res
        if self.test:
            test_res = self._named_metrics2res(self.test_named_metrics)
            dict2json(apply_format_to_float(f'{metric_path}/test.json', self.fmt), test_res)
            res['test'] = test_res
        if self.early_stop is not None:
            dict2json(f'{metric_path}/early_stop.json', self.early_stop.to_dict())

        return self._simplify_res(res)

    def _do_fold(
        self, 
        fold: Optional[int], 
        train_loader: Optional[DataLoader] = None, 
        val_loader: Optional[DataLoader] = None, 
        test_loader: Optional[DataLoader] = None,
        predict_loader: Optional[DataLoader] = None,
    ) -> None:
        """Process one fold of task.

        One fold can be composed of some of the 4 stages of {train, val, test, predict}.
        The usual order is: train->val->train->val->...->train->val->test->predict.
        """
        self.train_named_metrics.append({})
        self.val_named_metrics.append({})
        self.test_named_metrics.append({})

        if train_loader:
            for epoch in range(0 if self.val_first else 1, self.early_stop.max_stop_epoch + 1):
                if epoch > 0:
                    train_named_metrics = self._do_epoch('train', train_loader, fold, epoch)
                    if self.save_ckpt_period > 0 and epoch > 0 and epoch % self.save_ckpt_period == 0:
                        self._save_ckpt(fold, epoch)

                if val_loader:
                    val_named_metrics = self._do_epoch('val', val_loader, fold, epoch)

                    var_name, val_name = self.early_stop.metric_name.split('.')
                    self.early_stop(getattr(val_named_metrics[var_name], val_name), epoch)
                    if self.early_stop.best_epoch == epoch:
                        self._save_ckpt(fold, None)
                        self.val_named_metrics[-1] = val_named_metrics
                        if epoch > 0:
                            self.train_named_metrics[-1] = train_named_metrics

                    if self.early_stop.early_stop:
                        break
        else:
            if val_loader:
                self.val_named_metrics[-1] = self._do_epoch('val', val_loader, fold, -1)

        if train_loader:
            self._load_ckpt(fold, None)
        if test_loader:
            self.test_named_metrics[-1] = self._do_epoch('test', test_loader, fold, -1)
        if predict_loader:
            self._do_epoch('predict', predict_loader, fold, -1)

    def _do_epoch(
        self,
        stage: str,
        loader: Optional[DataLoader],
        fold: int,
        epoch: int
    ) -> Optional[dict[str, Metric]]:
        """Process one iteration of `DataLoader`"""
        env.stage, env.fold, env.epoch = stage, fold, epoch
        if loader is None:
            return None

        with WithNone() if stage == 'train' else torch.no_grad(), EpochProgress(stage, len(loader), fold, epoch) as progress:
            getattr(self.model_module, f'{stage}_epoch_begin')()

            for model in self.model_module.named_models.values():
                model.train() if stage == 'train' else model.eval()

            for step, batch in enumerate(loader, 1):
                step_res = self._do_step(stage, step, batch) or {}
                progress.update(**step_res)

                self._update_tensorboard(stage, self.model_module.named_metrics, step=epoch * len(loader) + step)
            self._update_tensorboard(stage, self.model_module.named_metrics, epoch=epoch)

            getattr(self.model_module, f'{stage}_epoch_end')()

            named_metrics = copy.deepcopy(self.model_module.named_metrics)
            for metric in self.model_module.named_metrics.values():
                metric.reset()
            return named_metrics

    def _do_step(
        self, 
        stage: str, 
        step: int, 
        batch,
    ) -> None:
        """Process one batch of data."""
        env.step = step
        with autocast():
            batch = self._batch_to_cuda(batch)
            return getattr(self.model_module, f'{stage}_step')(batch)

    def _save_ckpt(
        self, 
        fold: Optional[int] = None, 
        epoch: Optional[int] = None
    ) -> None:
        """Save checkpoints to path f'{env.exp_path}/ckpt/{fold}'.

        You can rewrite `yangdl.ModelModule.save_ckpt` function to decide exactly what to save.
        """
        ckpt_path = f'{env.exp_path}/ckpt'
        if fold is not None:
            ckpt_path = os.path.join(ckpt_path, str(fold))
        os.makedirs(ckpt_path, exist_ok=True)
        ckpt_name = os.path.join(ckpt_path, 'best.pt' if epoch is None else f'{epoch}.pt')

        self.model_module.save_ckpt(ckpt_name)

    def _load_ckpt(
        self, 
        fold: Optional[int] = None, 
        epoch: Optional[int] = None
    ) -> None:
        """Load checkpoints from path f'{env.exp_path}/ckpt/{fold}'.

        It is mainly used to load the best model during the test stage.
        """
        ckpt_path = f'{env.exp_path}/ckpt'
        if fold is not None:
            ckpt_path = os.path.join(ckpt_path, str(fold))
        ckpt_name = os.path.join(ckpt_path, 'best.pt' if epoch is None else f'{epoch}.pt')

        self.model_module.load_ckpt(ckpt_name)

    def _named_metrics2res(self, named_metrics: list[dict[str, Metric]]) -> dict:
        """Convert named_metrics to a combination of dict and list.

        Returns:
            res example: {'loss': {'loss': {'data': [0.1, 0.3], 'mean': 0.2, 'std': 0.01}}}.
        """
        res = defaultdict(lambda: defaultdict(lambda: {'data': [], 'mean': None, 'std': None}))
        for named_metric in named_metrics:
            for metric_name, metric in named_metric.items():
                for prop_name, prop in metric.data.items():
                    res[metric_name][prop_name]['data'].append(prop)

        for metric_dict in res.values():
            for prop_dict in metric_dict.values():
                arr = np.array(prop_dict['data'])
                prop_dict['mean'] = arr.mean(axis=0).tolist()
                prop_dict['std'] = arr.std(axis=0).tolist()

        return res
    
    def _simplify_res(self, res: dict) -> dict:
        """Remove res's 'data' and 'std', only keep 'mean'. Because people care more about the mean value of metrics."""
        res = copy.deepcopy(res)
        for _, stage_res in res.items():
            for metric_dict in stage_res.values():
                for prop_name, prop_dict in metric_dict.items():
                    metric_dict[prop_name] = prop_dict['mean']

        return res

    def _update_tensorboard(
        self,
        stage: str,
        named_metrics: dict[str, Metric],
        epoch: Optional[int] = None,
        step: Optional[int] = None
    ):
        """Log the scalar value of named_metrics to tensorboard."""
        if epoch == -1:
            return
        for metric in named_metrics.values():
            if epoch is not None and metric.freq[0] == 'epoch' and epoch % metric.freq[1] == 0:
                self.tensorboard_logger.log(metric, stage, epoch)
            if step is not None and metric.freq[0] == 'step' and step % metric.freq[1] == 0:
                self.tensorboard_logger.log(metric, stage, step)
    
    def _batch_to_cuda(self, batch):
        """Move cpu data in batch to gpu.

        TODO: Refactor this part.
        """
        if isinstance(batch, Tensor):
            batch = batch.cuda()
        elif isinstance(batch, dict):
            for key, val in batch.items():
                if isinstance(val, Tensor):
                    batch[key] = val.cuda()
        elif isinstance(batch, list):
            if isinstance(batch[0], (list, tuple)):
                batch = list(list(map(lambda x: x.cuda() if isinstance(x, Tensor) else x, item)) for item in batch)
            else:
                batch = [(item.cuda() if isinstance(item, Tensor) else item) for item in batch]
        else:
            raise Exception

        return batch




import os
import random
import shutil
from typing import Optional

import numpy as np
import torch

from yangdl.utils.python import get_caller_file_name


__all__ = [
    'env',
]


class Env:
    """Environmental context of `yangdl` (Use singleton `env` instead of instantiating `Env`).

    Used to initialize `yangdl` configuration and manage global variables.
    """
    # set these manually
    exp_path: str = None  # experiment directory. Set to `None` generally for `predict` task.
    seed: int = None  # it is best to set seed for reproduction

    # Automatically generated values:
    stage: str = None # Value in {'train', 'val', 'test', 'predict'}.
    fold: int = None # The count of current fold.
    epoch: int = None # The count of current epoch.
    step: int = None # The count of current step.

    def __setattr__(self, key, val):
        self.__dict__[key] = val

        if key == 'exp_path':
            for file_name in ('log', 'metric', 'ckpt'):
                os.makedirs(f'{val}/{file_name}', exist_ok=True)
            shutil.copyfile(get_caller_file_name(), f'{val}/code.py')

            from yangdl.core.logger import logger
            logger.handler.set_file_name(f'{val}/log/log.txt')
        elif key == 'seed':
            self._set_seed(val)

    def __getattr__(self, key):
        return None

    @staticmethod
    def _set_seed(seed: Optional[int] = None):
        if seed is None:
            return
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed) # if you are using multi-GPU.
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True


env = Env()




from yangdl.core import (
    ModelModule,
    DataModule,
    TaskModule,
    logger,
    env,
)
from yangdl.metric import (
    Metric,
    ValueMetric,
    ClsMetric,
    SegMetric,
    RegMetric,
)
from yangdl import utils


__all__ = [
    'ModelModule',
    'DataModule',
    'TaskModule',
    'logger',
    'env',
    'Metric',
    'ValueMetric',
    'ClsMetric',
    'SegMetric',
    'RegMetric',
    'utils',
]



import importlib
import inspect
import os
from typing import Callable


def get_properties(obj) -> dict:
    return {key: obj.__getattribute__(key) for key, val in vars(obj.__class__).items() if isinstance(val, property)}


def get_class_name_from_method(f: Callable):
    return f.__qualname__.split('.')[0]


def get_class_from_method(f: Callable):
    cls_name = get_class_name_from_method(f)
    return getattr(importlib.import_module(f.__module__), cls_name)


def method_is_overrided_in_subclass(f: Callable):
    cls = get_class_from_method(f)
    for parent in cls.__mro__[1:]:
        if f.__name__ in parent.__dict__:
            return True
    return False


def get_caller_file_name():
    frame_info = inspect.stack()[-1]
    caller_file = frame_info.filename

    return os.path.abspath(caller_file)




import torch
from torch import Tensor
from sklearn.metrics import roc_auc_score, roc_curve, precision_recall_curve, average_precision_score
from typing import Optional

from yangdl.metric import Metric


__all__ = [
    'ConfusionMatrix',
]


class ConfusionMatrix(Metric):
    def __init__(
        self, 
        num_classes: int,
        properties: Optional[list[str]] = None,
        thresh: Optional[float | str] = None, 
        ignore_label: Optional[int] = None,
        eps: float = 1e-7,
        **kwargs,
    ):
        """
        Args:
            num_classes: The number of classes.
                If `num_classes == 2`, the negative class(label is equal to 0) does not participate in the metric calculation.
            properties: The properties you are interested in will be
                automatically logged in tensorboard and printed and saved as results.
                If is None, will save all properties
                (check the code to see what properties are there).
            thresh: The threshold that determines whether a sample is positive or negative.
                Only useful in binary classification.
                If is float, the value range is between 0 and 1.
                If is 'f1_score', will use threshold that has best f1_score.
                If is 'roc', will use threshold that has best tpr - fpr.
            ignore_label: Value in [0, C], this label will not use to calculate metrics.
            eps: Prevent division by 0.
        """
        super().__init__(**kwargs)

        self.num_classes = num_classes
        self.properties = properties
        if thresh is None and num_classes == 2:
            thresh = 0.5
        self._thresh = thresh
        self.ignore_label = ignore_label

        self.add_tensor('_matrix', torch.zeros((num_classes, num_classes), dtype=torch.long))
        self.eps = eps
        self.save_probs = 'auc' in properties or 'ap' in properties or isinstance(thresh, str)

        if self.save_probs:
            self.add_tensor('probs', torch.tensor([], dtype=torch.float)) # num_classes == 2: (B,), else: (B, C)
            self.add_tensor('labels', torch.tensor([], dtype=torch.long)) # (B,)

    @torch.no_grad()
    def update(
        self,
        probs: Tensor,
        labels: Tensor,
    ) -> None:
        """Update metric with probabilities and labels.

        Args:
            probs: probability predicted by the model.
                Shape is (B,) or (B, C). Value is in [0, 1].
            labels: ground truth.
                Shape is (B,). Value is in (0, 1, ..., C - 1).
        """
        probs = probs.detach()
        if self.num_classes == 2 and len(probs.shape) > len(labels.shape):
            probs = probs[:, -1]
        if self.ignore_label is not None:
            mask = labels != self.ignore_label
            probs = probs[mask]
            labels = labels[mask]

        if not isinstance(self._thresh, str):
            self._matrix += self._calc_matrix(probs, labels)

        if self.save_probs:
            self.probs = torch.cat([self.probs, probs], dim=0)
            self.labels = torch.cat([self.labels, labels], dim=0)

    def _calc_matrix(
        self,
        probs: Tensor,
        labels: Tensor,
    ) -> Tensor:
        preds = self._calc_preds(probs)
        bins = self._bincount(labels * self.num_classes + preds, minl=self.num_classes ** 2)
        return bins.reshape(self.num_classes, self.num_classes)
    
    def _calc_preds(
        self,
        probs: Tensor,
    ) -> Tensor:
        if self.num_classes == 2:
            return (probs > self.thresh).long()
        else:
            return probs.argmax(dim=1).long()

    def _bincount(self, x: Tensor, minl: Optional[int] = None):
        """
        Example: x=[1, 1, 0, 2, 2] --> res=[1, 2, 2].
        """
        if minl is None:
            minl = x.max() + 1
        res = torch.empty(minl, device=x.device, dtype=torch.long)
        for i in range(minl):
            res[i] = (x == i).sum()
        return res

    def ravel(self, idx: Optional[int] = None):
        """Convert confusion matrix to (tn, fp, fn, tp).

        Args:
            idx: Positive class idx.
        """
        tp = self.matrix[idx, idx]
        fp = self.matrix[:, idx].sum() - tp
        fn = self.matrix[idx, :].sum() - tp
        tn = self.matrix.sum() - tp - fp - fn

        return tn, fp, fn, tp

    @property
    def matrix(self):
        if isinstance(self._thresh, str) and self._matrix.sum() != len(self.probs):
            self._matrix = self._calc_matrix(self.probs, self.labels)
        return self._matrix

    @property
    def acc(self):
        return self.matrix.diagonal().sum() / (self.matrix.sum() + self.eps)

    @property
    def precision(self):
        return self.precisions.mean()
    
    @property
    def precisions(self):
        precisions = []
        for idx in range(self.num_classes == 2, self.num_classes):
            tn, fp, fn, tp = self.ravel(idx)
            precisions.append(tp / (tp + fp + self.eps))
        return torch.stack(precisions, dim=0)
    
    @property
    def recall(self):
        return self.recalls.mean()
    
    @property
    def recalls(self):
        recalls = []
        for idx in range(self.num_classes == 2, self.num_classes):
            tn, fp, fn, tp = self.ravel(idx)
            recalls.append(tp / (tp + fn + self.eps))
        return torch.stack(recalls, dim=0)

    @property
    def sensitivity(self):
        return self.recall

    @property
    def sensitivitys(self):
        return self.recalls

    @property
    def specificity(self):
        return self.specificitys.mean()

    @property
    def specificitys(self):
        specificitys = []
        for idx in range(self.num_classes == 2, self.num_classes):
            tn, fp, fn, tp = self.ravel(idx)
            specificitys.append(tn / (tn + fp + self.eps))
        return torch.stack(specificitys, dim=0)

    @property
    def f1_score(self):
        return 2 * self.precision * self.recall / (self.precision + self.recall + self.eps)

    @property
    def f1_scores(self):
        f1_scores = []
        for precision, recall in zip(self.precisions, self.recalls):
            f1_scores.append(2 * precision * recall / (precision + recall + self.eps))
        return torch.stack(f1_scores, dim=0)

    @property
    def auc(self):
        assert self.save_probs is True
        res = -1.
        try:
            if self.num_classes == 2:
                if self.labels.sum().item() not in (0, len(self.labels)):
                    res = roc_auc_score(self.labels.cpu().numpy(), self.probs.cpu().numpy(), average='macro')
            else:
                res = roc_auc_score(self.labels.cpu().numpy(), self.probs.cpu().numpy(), average='macro', multi_class='ovr')
        except Exception as e:
            print(e)
        return torch.tensor(res, dtype=torch.float)

    @property
    def ap(self):
        assert self.save_probs is True
        assert self.num_classes == 2, f'num_classes={self.num_classes}'

        res = -1.
        if self.labels.sum().item() not in (0, len(self.labels)):
            res = average_precision_score(self.labels.cpu().numpy(), self.probs.cpu().numpy())
        return torch.tensor(res, dtype=torch.float)

    @property
    def dice(self):
        return self.dices.mean()
    
    @property
    def dices(self):
        dices = []
        for idx in range(self.num_classes == 2, self.num_classes):
            tn, fp, fn, tp = self.ravel(idx)
            dices.append(2 * tp / (fp + 2 * tp + fn + self.eps))
        return torch.stack(dices, dim=0)

    @property
    def iou(self):
        return self.ious.mean()

    @property
    def ious(self):
        ious = []
        for idx in range(self.num_classes == 2, self.num_classes):
            tn, fp, fn, tp = self.ravel(idx)
            ious.append(tp / (fp + tp + fn + self.eps))
        return torch.stack(ious, dim=0)
    
    @property
    def thresh(self):
        assert self.num_classes == 2, f'num_classes={self.num_classes}'

        if self._thresh is None:
            thresh = -1.

        if not isinstance(self._thresh, str):
            thresh = self._thresh
        else:
            assert self.save_probs is True
            labels, probs = self.labels.cpu().numpy(), self.probs.cpu().numpy()
            if labels.sum().item() in (0, len(labels)):
                thresh = 0.5
            else:
                if self._thresh == 'f1_score':
                    precisions, recalls, threshs = precision_recall_curve(labels, probs)
                    f1s = 2 * precisions * recalls / (precisions + recalls + self.eps)
                    thresh = threshs[f1s.argmax(axis=0)]
                elif self._thresh == 'roc':
                    fpr, tpr, threshs = roc_curve(labels, probs, pos_label=1)
                    thresh = threshs[(tpr - fpr).argmax(axis=0)]

        return torch.tensor(thresh, dtype=torch.float)




from copy import deepcopy
from typing import Optional

from torch import Tensor


class Metric():
    """Base metric class.

    Need to be inherited and implement `update` method.
    """
    def __init__(
        self,
        prefix: Optional[str] = None, 
        freq: tuple[str, int] = ('epoch', 1),
    ):
        """
        Args:
            prefix: If is `None`, tensorboard log will use var name 'train/auc'.
                If is `str`, tensorboard log will use var name f'train/{prefix}_auc'.
            freq: The frequency of logging to tensorboard.
                Examples: ('epoch', 1): log every epoch. ('step', 5): log every 5 steps.
        """
        # use for tensorboard logger
        self.prefix = prefix
        self.freq = freq
        self.properties = None

        self.named_tensors = {}

    def add_tensor(
        self, 
        name: str, 
        tensor: Tensor, 
    ) -> None:
        """Add automatically managed tensor.

        Args:
            name: variable name
            tensor: tensor default value
        """
        self.named_tensors[name] = tensor
        setattr(self, name, deepcopy(tensor))

    def cuda(self):
        """Move all tensor and Metric's tensor to cuda (Modified in-place.)."""
        self.named_tensors = {name: tensor.cuda() for name, tensor in self.named_tensors.items()}

        for name, val in vars(self).items():
            if isinstance(val, Tensor):
                setattr(self, name, val.cuda())
            if isinstance(val, Metric):
                val.cuda()

        return self

    def reset(self) -> None:
        """Reset all tensor and `Metric`."""
        for name, tensor in self.named_tensors.items():
            setattr(self, name, deepcopy(tensor))

        for name, val in vars(self).items():
            if isinstance(val, Metric):
                getattr(self, name).reset()

    @property
    def data(self) -> dict:
        """Return all {property_name: property} pair of this `Metric`."""
        return {key: getattr(self, key).cpu().tolist() for key in self.properties}




from typing import Optional, Sequence

import torch
from torch import Tensor

from yangdl.metric.confusion_matrix import ConfusionMatrix


__all__ = [
    'SegMetric',
]


class SegMetric(ConfusionMatrix):
    """The metric of segmentation task (See base class `ConfusionMatrix` for more information)."""
    def __init__(
        self, 
        num_classes: int,
        properties: Optional[Sequence[str]] = None,
        thresh: Optional[float | str] = None, 
        ignore_label: Optional[int] = None,
        eps: float = 1e-7, 
        **kwargs,
    ):
        """
        Args:
            num_classes: The number of classes.
            properties: The properties you are interested in will be
                automatically logged in tensorboard and printed and saved as results.
                If is `None`, will save all properties
                (check the code to see what properties are there).
            thresh: The threshold that determines whether a sample is positive or negative.
                Only useful in binary classification.
                If is `float`, the value must in [0, 1].
                If is 'f1_score', will use threshold that has best f1_score.
                If is 'roc', will use threshold that has best tpr - fpr.
            ignore_label: Value in [0, C], this label will not used to calculate metrics.
            eps: Prevent division by 0.
        """
        if properties is None:
            if num_classes == 2:
                properties = ['acc', 'precision', 'recall', 'sensitivity', 'specificity', 'f1_score', 'dice', 'iou', 'thresh']
            else:
                properties = ['acc', 'precision', 'recall', 'sensitivity', 'specificity', 'f1_score', 'dice', 'iou']
        super().__init__(num_classes, properties, thresh, ignore_label, eps, **kwargs)

    @torch.no_grad()
    def update(
        self,
        probs: Tensor,
        labels: Tensor,
    ):
        """Update metric with probabilities and labels.

        Args:
            probs: probability predicted by the model.
                Shape is (B, [D], H, W) or (B, C, [D], H, W). Value is in [0, 1].
            labels: ground truth.
                Shape is (B, [D], H, W). Value is in (0, 1, ..., C - 1).
        """
        probs = probs.detach()
        if len(probs.shape) > len(labels.shape):
            probs = probs.moveaxis(1, -1)
            probs = probs.flatten(0, -2)
        else:
            probs = probs.flatten(0, -1)
        labels = labels.flatten(0, -1)

        super().update(probs, labels)




import numpy as np


__all__ = [
    'EarlyStop',
]


class EarlyStop():
    """Early stop in `train` stage based on performance during the `val` stage."""
    def __init__(
        self,
        monitor: dict[str] = {'loss.val': 'small'},
        delta: float = 0.,
        patience: int = 10,
        min_stop_epoch: int = 10,
        max_stop_epoch: int = 100,
    ):
        """
        Args:
            monitor: Monitored metric used to decide whether to early stop.
                Examples: {'loss.loss': 'small'}: Model will early stop with the lowest loss.
                {'metric.auc': 'big'}: Model will early stop with the highest auc.
            delta: The performance will be considered better only if the result is at least `delta` better than the past.
            patience: If the performance is not better in the `val` stage for consecutive `patience` epochs,
                `train` will be stopped.
            min_stop_epoch: Will not early stop when epoch is less than `min_stop_epoch`.
            max_stop_epoch: Train will be forced to stop at `max_stop_epoch`.
        """
        self.metric_name, self.rule = next(iter(monitor.items()))
        self.delta = delta
        self.patience = patience
        self.min_stop_epoch = min_stop_epoch
        self.max_stop_epoch = max_stop_epoch
        self.res = {'stop_epoch': [], 'best_epoch': []}

        self.init()

    def init(self):
        self.counter = 0
        self.best_val = {'small': np.Inf, 'big': -np.Inf}[self.rule]
        self.stop_epoch = 0
        self.best_epoch = 0
        self.early_stop = False

    def __call__(self, val, epoch):
        if self.rule == 'small':
            if val < self.best_val - self.delta:
                self.best_val = val
                self.counter = 0
                self.best_epoch = epoch
            else:
                self.counter += 1
        elif self.rule == 'big':
            if val > self.best_val + self.delta:
                self.best_val = val
                self.counter = 0
                self.best_epoch = epoch
            else:
                self.counter += 1

        if self.counter > self.patience and epoch >= self.min_stop_epoch or epoch == self.max_stop_epoch:
            self.stop_epoch = epoch
            self.early_stop = True
            self.res['stop_epoch'].append(self.stop_epoch)
            self.res['best_epoch'].append(self.best_epoch)

    def to_dict(self):
        return self.res




from typing import Generator

from torch.utils.data import DataLoader


__all__ = [
    'DataModule',
]


class DataModule():
    """Generate `DataLoader` for each fold.

    You should inherit `DataModule` and override at least one method in
    `{train, val, test, predict}_loader` according to your own task.
    """
    def __init__(self):
        pass

    def train_loader(self) -> Generator[DataLoader, None, None]:
        while True:
            yield None

    def val_loader(self) -> Generator[DataLoader, None, None]:
        while True:
            yield None

    def test_loader(self) -> Generator[DataLoader, None, None]:
        while True:
            yield None

    def predict_loader(self) -> DataLoader:
        while True:
            yield None




from setuptools import (
    setup,
    find_packages,
)


setup(
    name='yangdl',
    version='0.1.0',
    packages=find_packages(),
    url='https://github.com/m1dsolo/yangdl',
    license='MIT',
    author='m1dsolo',
    author_email='yx1053532442@gmail.com',
    description='A simple pytorch-based framework for multi-fold train, val, test, predict!',
    long_description='https://github.com/m1dsolo/yangdl',
    install_requires=[
        'numpy',
        'scikit-learn',
        'rich',
        'tensorboard',
        'torch',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.10',
)



from yangdl.metric.metric import Metric
from yangdl.metric.value import ValueMetric
from yangdl.metric.cls import ClsMetric
from yangdl.metric.seg import SegMetric
from yangdl.metric.reg import RegMetric


__all__ = [
    'Metric',
    'ValueMetric',
    'ClsMetric',
    'SegMetric',
    'RegMetric',
]



import torch
from torch import Tensor

from yangdl.metric import Metric


class RegMetric(Metric):
    """The metric of regression task."""
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.add_tensor('preds', torch.tensor([], dtype=torch.float32))
        self.add_tensor('labels', torch.tensor([], dtype=torch.float32))

        self.properties = ['mae', 'rmse', 'r2']

    @torch.no_grad()
    def update(self, preds: Tensor, labels: Tensor):
        """Update metric with probabilities and labels.

        Args:
            preds: value predicted by the model. Shape is (B,).
                
            labels: ground truth. Shape is (B,)
        """
        assert len(preds) == len(labels) and len(preds.shape) == 1, f'preds.shape={preds.shape}, labels.shape={labels.shape}'

        preds, labels = preds.detach(), labels.detach()

        self.preds = torch.cat([self.preds, preds], dim=0)
        self.labels = torch.cat([self.labels, labels], dim=0)

    @property
    def mae(self) -> Tensor:
        return (self.labels - self.preds).abs().mean()

    @property
    def rmse(self) -> Tensor:
        return ((self.labels - self.preds) ** 2).mean().sqrt()

    @property
    def r2(self) -> Tensor:
        a = ((self.labels - self.labels.mean()) ** 2).sum()
        b = ((self.labels - self.preds) ** 2).sum()

        return 1 - b / a




import torch
from torch import nn
from torch.nn.modules import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from yangdl.metric import Metric


__all__ = [
    'ModelModule',
]


class ModelModule():
    """Encapsulate models and task implementation.

    You should inherit `ModelModule` and override at least one method in `{train, val, test, predict}_step` according to your own task.
    """
    def __init__(self):
        self._registered_models = {}

    def __iter__(self):
        """Called on each new fold.

        You don't need to yield anything, you mainly need to initialize a new model for each fold.
        """
        yield

    def train_step(self, batch) -> None:
        """Overridable train function called at each step."""
        pass

    def val_step(self, batch) -> None:
        """Overridable val function called at each step."""
        pass

    def test_step(self, batch) -> None:
        """Overridable test function called at each step."""
        pass

    def predict_step(self, batch) -> dict:
        """Overridable predict function called at each step."""
        pass

    def train_epoch_begin(self) -> None:
        """Overridable callback function at the begin of epoch in the train stage."""
        pass

    def val_epoch_begin(self) -> None:
        """Overridable callback function at the begin of epoch in the val stage."""
        pass

    def test_epoch_begin(self) -> None:
        """Overridable callback function at the begin of epoch in the test stage."""
        pass

    def predict_epoch_begin(self) -> None:
        """Overridable callback function at the begin of epoch in the predict stage."""
        pass

    def train_epoch_end(self) -> None:
        """Overridable callback function at the end of epoch in the train stage."""
        pass

    def val_epoch_end(self) -> None:
        """Overridable callback function at the end of epoch in the val stage."""
        pass

    def test_epoch_end(self) -> None:
        """Overridable callback function at the end of epoch in the test stage."""
        pass

    def predict_epoch_end(self) -> None:
        """Overridable callback function at the end of epoch in the predict stage."""
        pass
    
    def register_models(self, d) -> None:
        """
        Models that are not `nn.Module` subclasses must be registered.
        The registered models will be automatically moved to gpu.
        The model must have `cuda` method.
        If the `ModelModule.save_ckpt` and `ModelModule.load_ckpt` methods are not overriden,
        the model must have `state_dict` and `load_state_dict` methods.
        """
        self._registered_models.update(d)

    @property
    def named_models(self) -> dict[str, nn.Module]:
        return dict(filter(lambda kv: Module in kv[1].__class__.__mro__, self.__dict__.items())) | self._registered_models

    @property
    def named_metrics(self) -> dict[str, Metric]:
        return dict(filter(lambda kv: Metric in kv[1].__class__.__mro__, self.__dict__.items()))

    @property
    def named_optimizers(self) -> dict[str, Optimizer]:
        return dict(filter(lambda kv: Optimizer in kv[1].__class__.__mro__, self.__dict__.items()))

    @property
    def named_schedulers(self) -> dict[str, _LRScheduler]:
        return dict(filter(lambda kv: _LRScheduler in kv[1].__class__.__mro__, self.__dict__.items()))

    def cuda(self):
        """Move all models and metrics in `ModelModule` to gpu.This method modifies `ModelModule` in-place.
        """
        for model in self.named_models.values():
            model.cuda()
        for metric in self.named_metrics.values():
            metric.cuda()

        return self

    def save_ckpt(self, ckpt_name: str):
        """Save all models and optimizers' state_dict to ckpt_name.

        TODO: save lr_schedulers.
        """
        ckpt = {}
        for name, model in self.named_models.items():
            ckpt[name] = model.state_dict()
        for name, optimizer in self.named_optimizers.items():
            ckpt[name] = optimizer.state_dict()

        torch.save(ckpt, ckpt_name)

    def load_ckpt(self, ckpt_name: str):
        """Load all models and optimizers' state_dict from ckpt_name."""
        ckpt = torch.load(ckpt_name)
        for name, state_dict in ckpt.items():
            getattr(self, name).load_state_dict(state_dict)

    @property
    def device(self) -> torch.device:
        """
        Currently the framework only supports a single gpu.
        So this method will only return a single device.
        """
        for model in self.named_models.values():
            for param in model.parameters():
                return param.device

